@@ -271,6 +271,12 @@ static const char *commands[NL80211_CMD_MAX + 1] = {
 	[NL80211_CMD_DEL_NAN_FUNCTION] = "del_nan_function",
 	[NL80211_CMD_CHANGE_NAN_CONFIG] = "change_nan_config",
 	[NL80211_CMD_NAN_MATCH] = "nan_match",
+	[NL80211_CMD_SET_MULTICAST_TO_UNICAST] = "set_multicast_to_unicast",
+	[NL80211_CMD_UPDATE_CONNECT_PARAMS] = "update_connect_params",
+	[NL80211_CMD_SET_PMK] = "set_pmk",
+	[NL80211_CMD_DEL_PMK] = "del_pmk",
+	[NL80211_CMD_PORT_AUTHORIZED] = "port_authorized",
+	[NL80211_CMD_RELOAD_REGDB] = "reload_regdb",
 };
 
 static char cmdbuf[100];
@@ -366,7 +372,7 @@ static int hex2byte(const char *hex)
 	return (d1 << 4) | d2;
 }
 
-static char *hex2bin(const char *hex, char *buf)
+char *hex2bin(const char *hex, char *buf)
 {
 	char *result = buf;
 	int d;
@@ -383,16 +389,101 @@ static char *hex2bin(const char *hex, char *buf)
 	return result;
 }
 
+static int parse_akm_suite(const char *cipher_str)
+{
+
+	if (!strcmp(cipher_str, "PSK"))
+		return 0x000FAC02;
+	if (!strcmp(cipher_str, "FT/PSK"))
+		return 0x000FAC03;
+	if (!strcmp(cipher_str, "PSK/SHA-256"))
+		return 0x000FAC06;
+	return -EINVAL;
+}
+
+static int parse_cipher_suite(const char *cipher_str)
+{
+
+	if (!strcmp(cipher_str, "TKIP"))
+		return 0x000FAC02;
+	if (!strcmp(cipher_str, "CCMP") || !strcmp(cipher_str, "CCMP-128"))
+		return 0x000FAC04;
+	if (!strcmp(cipher_str, "GCMP") || !strcmp(cipher_str, "GCMP-128"))
+		return 0x000FAC08;
+	if (!strcmp(cipher_str, "GCMP-256"))
+		return 0x000FAC09;
+	if (!strcmp(cipher_str, "CCMP-256"))
+		return 0x000FAC0A;
+	return -EINVAL;
+}
+
 int parse_keys(struct nl_msg *msg, char **argv, int argc)
 {
 	struct nlattr *keys;
 	int i = 0;
 	bool have_default = false;
+	char *arg = *argv;
 	char keybuf[13];
+	int pos = 0;
 
 	if (!argc)
 		return 1;
 
+	if (!memcmp(&arg[pos], "psk", 3)) {
+		char psk_keybuf[32];
+		int cipher_suite, akm_suite;
+
+		if (argc < 4)
+			goto explain;
+
+		pos+=3;
+		if (arg[pos] != ':')
+			goto explain;
+		pos++;
+
+		NLA_PUT_U32(msg, NL80211_ATTR_WPA_VERSIONS, NL80211_WPA_VERSION_2);
+
+		if (strlen(&arg[pos]) != (sizeof(psk_keybuf) * 2) || !hex2bin(&arg[pos], psk_keybuf)) {
+			printf("Bad PSK\n");
+			return -EINVAL;
+		}
+
+		NLA_PUT(msg, NL80211_ATTR_PMK, 32, psk_keybuf);
+		NLA_PUT_U32(msg, NL80211_ATTR_AUTH_TYPE, NL80211_AUTHTYPE_OPEN_SYSTEM);
+
+		argv++;
+		argc--;
+		arg = *argv;
+
+		akm_suite = parse_akm_suite(arg);
+		if (akm_suite < 0)
+			goto explain;
+
+		NLA_PUT_U32(msg, NL80211_ATTR_AKM_SUITES, akm_suite);
+
+		argv++;
+		argc--;
+		arg = *argv;
+
+		cipher_suite = parse_cipher_suite(arg);
+		if (cipher_suite < 0)
+			goto explain;
+
+		NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITES_PAIRWISE, cipher_suite);
+
+		argv++;
+		argc--;
+		arg = *argv;
+
+		cipher_suite = parse_cipher_suite(arg);
+		if (cipher_suite < 0)
+			goto explain;
+
+		NLA_PUT_U32(msg, NL80211_ATTR_CIPHER_SUITE_GROUP, cipher_suite);
+
+		return 0;
+	}
+
 	NLA_PUT_FLAG(msg, NL80211_ATTR_PRIVACY);
 
 	keys = nla_nest_start(msg, NL80211_ATTR_KEYS);
@@ -400,11 +491,13 @@ int parse_keys(struct nl_msg *msg, char **argv, int argc)
 		return -ENOBUFS;
 
 	do {
-		char *arg = *argv;
-		int pos = 0, keylen;
+		int keylen;
 		struct nlattr *key = nla_nest_start(msg, ++i);
 		char *keydata;
 
+		arg = *argv;
+		pos = 0;
+
 		if (!key)
 			return -ENOBUFS;
 
@@ -465,10 +558,284 @@ int parse_keys(struct nl_msg *msg, char **argv, int argc)
 			"  'index:' is a single digit (0-3)\n"
 			"  'data'   must be 5 or 13 ascii chars\n"
 			"           or 10 or 26 hex digits\n"
-			"for example: d:2:6162636465 is the same as d:2:abcde\n");
+			"for example: d:2:6162636465 is the same as d:2:abcde\n"
+			"or psk:data <AKM Suite> <pairwise CIPHER> <groupwise CIPHER> where\n"
+			"  'data' is the PSK (output of wpa_passphrase and the CIPHER can be CCMP or GCMP\n"
+			"for example: psk:0123456789abcdef PSK CCMP CCMP\n"
+			"The allowed AKM suites are PSK, FT/PSK, PSK/SHA-256\n"
+			"The allowed Cipher suites are TKIP, CCMP, GCMP, GCMP-256, CCMP-256\n");
 	return 2;
 }
 
+enum nl80211_chan_width str_to_bw(const char *str)
+{
+	static const struct {
+		const char *name;
+		unsigned int val;
+	} bwmap[] = {
+		{ .name = "5", .val = NL80211_CHAN_WIDTH_5, },
+		{ .name = "10", .val = NL80211_CHAN_WIDTH_10, },
+		{ .name = "20", .val = NL80211_CHAN_WIDTH_20, },
+		{ .name = "40", .val = NL80211_CHAN_WIDTH_40, },
+		{ .name = "80", .val = NL80211_CHAN_WIDTH_80, },
+		{ .name = "80+80", .val = NL80211_CHAN_WIDTH_80P80, },
+		{ .name = "160", .val = NL80211_CHAN_WIDTH_160, },
+	};
+	unsigned int i;
+
+	for (i = 0; i < ARRAY_SIZE(bwmap); i++) {
+		if (strcasecmp(bwmap[i].name, str) == 0)
+			return bwmap[i].val;
+	}
+
+	return NL80211_CHAN_WIDTH_20_NOHT;
+}
+
+static int parse_freqs(struct chandef *chandef, int argc, char **argv,
+		       int *parsed)
+{
+	uint32_t freq;
+	char *end;
+	bool need_cf1 = false, need_cf2 = false;
+
+	if (argc < 1)
+		return 0;
+
+	chandef->width = str_to_bw(argv[0]);
+
+	switch (chandef->width) {
+	case NL80211_CHAN_WIDTH_20_NOHT:
+		/* First argument was not understood, give up gracefully. */
+		return 0;
+	case NL80211_CHAN_WIDTH_20:
+	case NL80211_CHAN_WIDTH_5:
+	case NL80211_CHAN_WIDTH_10:
+		break;
+	case NL80211_CHAN_WIDTH_80P80:
+		need_cf2 = true;
+		/* fall through */
+	case NL80211_CHAN_WIDTH_40:
+	case NL80211_CHAN_WIDTH_80:
+	case NL80211_CHAN_WIDTH_160:
+		need_cf1 = true;
+		break;
+	}
+
+	*parsed += 1;
+
+	if (!need_cf1)
+		return 0;
+
+	if (argc < 2)
+		return 1;
+
+	/* center freq 1 */
+	if (!*argv[1])
+		return 1;
+	freq = strtoul(argv[1], &end, 10);
+	if (*end)
+		return 1;
+	*parsed += 1;
+
+	chandef->center_freq1 = freq;
+
+	if (!need_cf2)
+		return 0;
+
+	if (argc < 3)
+		return 1;
+
+	/* center freq 2 */
+	if (!*argv[2])
+		return 1;
+	freq = strtoul(argv[2], &end, 10);
+	if (*end)
+		return 1;
+	chandef->center_freq2 = freq;
+
+	*parsed += 1;
+
+	return 0;
+}
+
+
+/**
+ * parse_freqchan - Parse frequency or channel definition
+ *
+ * @chandef: chandef structure to be filled in
+ * @chan: Boolean whether to parse a channel or frequency based specifier
+ * @argc: Number of arguments
+ * @argv: Array of string arguments
+ * @parsed: Pointer to return the number of used arguments, or NULL to error
+ *          out if any argument is left unused.
+ *
+ * The given chandef structure will be filled in from the command line
+ * arguments. argc/argv will be updated so that further arguments from the
+ * command line can be parsed.
+ *
+ * Note that despite the fact that the function knows how many center freqs
+ * are needed, there's an ambiguity if the next argument after this is an
+ * integer argument, since the valid channel width values are interpreted
+ * as such, rather than a following argument. This can be avoided by the
+ * user by giving "NOHT" instead.
+ *
+ * The working specifier if chan is set are:
+ *   <channel> [NOHT|HT20|HT40+|HT40-|5MHz|10MHz|80MHz]
+ *
+ * And if frequency is set:
+ *   <freq> [NOHT|HT20|HT40+|HT40-|5MHz|10MHz|80MHz]
+ *   <control freq> [5|10|20|40|80|80+80|160] [<center1_freq> [<center2_freq>]]
+ *
+ * If the mode/channel width is not given the NOHT is assumed.
+ *
+ * Return: Number of used arguments, zero or negative error number otherwise
+ */
+int parse_freqchan(struct chandef *chandef, bool chan, int argc, char **argv,
+		   int *parsed)
+{
+	char *end;
+	static const struct chanmode chanmode[] = {
+		{ .name = "HT20",
+		  .width = NL80211_CHAN_WIDTH_20,
+		  .freq1_diff = 0,
+		  .chantype = NL80211_CHAN_HT20 },
+		{ .name = "HT40+",
+		  .width = NL80211_CHAN_WIDTH_40,
+		  .freq1_diff = 10,
+		  .chantype = NL80211_CHAN_HT40PLUS },
+		{ .name = "HT40-",
+		  .width = NL80211_CHAN_WIDTH_40,
+		  .freq1_diff = -10,
+		  .chantype = NL80211_CHAN_HT40MINUS },
+		{ .name = "NOHT",
+		  .width = NL80211_CHAN_WIDTH_20_NOHT,
+		  .freq1_diff = 0,
+		  .chantype = NL80211_CHAN_NO_HT },
+		{ .name = "5MHz",
+		  .width = NL80211_CHAN_WIDTH_5,
+		  .freq1_diff = 0,
+		  .chantype = -1 },
+		{ .name = "10MHz",
+		  .width = NL80211_CHAN_WIDTH_10,
+		  .freq1_diff = 0,
+		  .chantype = -1 },
+		{ .name = "80MHz",
+		  .width = NL80211_CHAN_WIDTH_80,
+		  .freq1_diff = 0,
+		  .chantype = -1 },
+	};
+	const struct chanmode *chanmode_selected = NULL;
+	unsigned int freq;
+	unsigned int i;
+	int _parsed = 0;
+	int res = 0;
+
+	if (argc < 1)
+		return 1;
+
+	if (!argv[0])
+		goto out;
+	freq = strtoul(argv[0], &end, 10);
+	if (*end) {
+		res = 1;
+		goto out;
+	}
+
+	_parsed += 1;
+
+	memset(chandef, 0, sizeof(struct chandef));
+
+	if (chan) {
+		enum nl80211_band band;
+
+		band = freq <= 14 ? NL80211_BAND_2GHZ : NL80211_BAND_5GHZ;
+		freq = ieee80211_channel_to_frequency(freq, band);
+	}
+	chandef->control_freq = freq;
+	/* Assume 20MHz NOHT channel for now. */
+	chandef->center_freq1 = freq;
+
+	/* Try to parse HT mode definitions */
+	if (argc > 1) {
+		for (i = 0; i < ARRAY_SIZE(chanmode); i++) {
+			if (strcasecmp(chanmode[i].name, argv[1]) == 0) {
+				chanmode_selected = &chanmode[i];
+				_parsed += 1;
+				break;
+			}
+		}
+	}
+
+	/* channel mode given, use it and return. */
+	if (chanmode_selected) {
+		chandef->center_freq1 = get_cf1(chanmode_selected, freq);
+		chandef->width = chanmode_selected->width;
+		goto out;
+	}
+
+	/* This was a only a channel definition, nothing further may follow. */
+	if (chan)
+		goto out;
+
+	res = parse_freqs(chandef, argc - 1, argv + 1, &_parsed);
+
+ out:
+	/* Error out if parsed is NULL. */
+	if (!parsed && _parsed != argc)
+		return 1;
+
+	if (parsed)
+		*parsed = _parsed;
+
+	return res;
+}
+
+int put_chandef(struct nl_msg *msg, struct chandef *chandef)
+{
+	NLA_PUT_U32(msg, NL80211_ATTR_WIPHY_FREQ, chandef->control_freq);
+	NLA_PUT_U32(msg, NL80211_ATTR_CHANNEL_WIDTH, chandef->width);
+
+	switch (chandef->width) {
+	case NL80211_CHAN_WIDTH_20_NOHT:
+		NLA_PUT_U32(msg,
+			    NL80211_ATTR_WIPHY_CHANNEL_TYPE,
+			    NL80211_CHAN_NO_HT);
+		break;
+	case NL80211_CHAN_WIDTH_20:
+		NLA_PUT_U32(msg,
+			    NL80211_ATTR_WIPHY_CHANNEL_TYPE,
+			    NL80211_CHAN_HT20);
+		break;
+	case NL80211_CHAN_WIDTH_40:
+		if (chandef->control_freq > chandef->center_freq1)
+			NLA_PUT_U32(msg,
+				    NL80211_ATTR_WIPHY_CHANNEL_TYPE,
+				    NL80211_CHAN_HT40MINUS);
+		else
+			NLA_PUT_U32(msg,
+				    NL80211_ATTR_WIPHY_CHANNEL_TYPE,
+				    NL80211_CHAN_HT40PLUS);
+		break;
+	default:
+		break;
+	}
+
+	if (chandef->center_freq1)
+		NLA_PUT_U32(msg,
+			    NL80211_ATTR_CENTER_FREQ1,
+			    chandef->center_freq1);
+
+	if (chandef->center_freq2)
+		NLA_PUT_U32(msg,
+			    NL80211_ATTR_CENTER_FREQ2,
+			    chandef->center_freq2);
+
+	return 0;
+
+ nla_put_failure:
+	return -ENOBUFS;
+}
+
 static void print_mcs_index(const __u8 *mcs)
 {
 	int mcs_bit, prev_bit = -2, prev_cont = 0;
